#version 140

#ifdef GL_ES
precision highp float;
#endif

#include "../include/half_lambert.frag"
#include "../include/clip_plane.frag"

out vec4 fragColor;

// define this for cascade debug colors
//#define CASCADE_DEBUG


#ifndef NO_PROJECTED_TEXTURE
// projected spotlight texture
uniform sampler2D tex;
#endif
// deferred inputs
// deferred_diffuse
uniform sampler2D tex2;

#ifndef NO_NORM_BUFFER
// deferred_normal
uniform sampler2D tex3;
#endif

#ifndef NO_SPEC_BUFFER
// deferred_specular
uniform sampler2D tex4;
#endif

// depth
uniform sampler2D tex5;

#ifdef USE_AMBIENT
uniform float ambientScale;
uniform samplerCube tex6;
#endif

// xy = positive 1/screen dims
// zw = projection matrix components
uniform vec4 ofs;

// these two specify frustum corners
uniform vec3 corner;
uniform vec3 cornerDelta;

uniform vec3 light_pos_view;
// x = radius, y = inner radius, z = outer cone cos (spot), w = inner cone cos (spot)
uniform vec4 light_radius;
uniform vec4 light_color;

uniform vec3 spot_dir_view[6];

uniform mat3 world_to_view;

uniform mat4 spotlight_projection_matrix[6];
// used to adjust bias (ortho only!)
uniform float spotlight_texel_size[6];

#ifdef USE_CASCADES
uniform int num_cascades;
#endif

#ifdef USE_SHADOWMAP
// what about transforming to get non-linear depth in shadowmap space?
// alternatively I might extract linear depth from shadowmap and project to view space;
// this seems like an overkill however
uniform
#	ifdef GL_ES
highp
#	endif
sampler2DShadow tex8
#if defined(OMNI) || defined(USE_CASCADES)
[6];
#else
[1];
#endif

// 1.0 = +-1 texel
#define PCF_SAMPLE_SCALE 1.0

// use 16-sample PCF without dithering, nice but slower
//#define PCF_16

// number of PCF samples to use with HW 2x2 PCF, we can afford less samples than old 8
#ifndef PCF_16
#	define PCF_SAMPLES 5
#else
#	define PCF_SAMPLES 16
#endif

#endif

float LinearDepth(float d)
{
	// now with pre-baked denormalization and sign flip
	return ofs.w / (ofs.z + d);
}

#ifdef USE_CASCADES

float compute_cascade_blend(vec2 uv)
{
	// 10% blend range by default
	float range = 0.1;
	float omrange = 1.0 - range;

	uv -= vec2(0.5);
	uv *= 2.0;

	uv = abs(uv);

	float blendx = 0.0;
	float blendy = 0.0;

	if (uv.x > omrange)
		blendx = (uv.x - omrange) * (1.0 / range);

	if (uv.y > omrange)
		blendy = (uv.y - omrange) * (1.0 / range);

	return max(blendx, blendy);
}

#endif

void main()
{
	clip_pixel();

#ifdef USE_SHADOWMAP
	vec2 samples[PCF_SAMPLES]=vec2[PCF_SAMPLES](
#ifndef PCF_16
	vec2(-0.973062, -0.328256),
	vec2(0.110481, 0.712614),
	vec2(-0.818603, 0.638871),
	vec2(-0.018864, -0.385902),
	vec2(-0.430052, 0.142593)
#else
	vec2(0.47914, 0.214901),
	vec2(-0.458871, -0.738133),
	vec2(-0.45404, 0.153784),
	vec2(0.473617, -0.72407),
	vec2(0.961697, 0.770339),
	vec2(-0.984345, -0.219209),
	vec2(0.034841, 0.831783),
	vec2(-0.014436, -0.163315),
	vec2(0.452166, -0.254094),
	vec2(-0.465602, 0.725876),
	vec2(0.038325, 0.292436),
	vec2(-0.972486, -0.710177),
	vec2(-0.908063, 0.252674),
	vec2(0.536454, 0.796915),
	vec2(-0.556828, -0.342318),
	vec2(-0.005866, -0.55867)
#endif
	);

#	ifndef PCF_16
	mat4 dither = mat4(
	0.0/16.0-0.5, 8.0/16.0-0.5, 2.0/16.0-0.5, 10.0/16.0-0.5,
	12.0/16.0-0.5, 4.0/16.0-0.5, 14.0/16.0-0.5, 6.0/16.0-0.5,
	3.0/16.0-0.5, 11.0/16.0-0.5, 1.0/16.0-0.5, 9.0/16.0-0.5,
	15.0/16.0-0.5, 7.0/16.0-0.5, 13.0/16.0-0.5, 5.0/16.0-0.5);
#	endif
#endif

	vec2 uv = gl_FragCoord.xy * ofs.xy;
	float storedDepth = texture(tex5, uv).x;

	vec3 diffuse = texture(tex2, uv).xyz;

#ifndef NO_NORM_BUFFER
	// unpack world space normal
	vec3 normal = texture(tex3, uv).xyz;

	// we store 0 into normal buffer for sky, it's an invalid value so this check gives a small speedup
	// (especially with shadow mapping, if a large portion of sky is visible)
	if (dot(normal, normal) != 0.0)
	{
		normal *= 2.0;
		normal -= vec3(1.0);
		// renormalization is necessary to reduce quantization errors
		normal = normalize(normal);
	}
#endif

#ifndef NO_SPEC_BUFFER
	vec4 specular = texture(tex4, uv);
#endif

	vec3 viewPos = vec3(0.0);

	float linear_depth;

	linear_depth = LinearDepth(storedDepth);
	viewPos = (corner + gl_FragCoord.xyz*cornerDelta)*linear_depth;

	vec3 light_delta = light_pos_view - viewPos;

	float dist = length(light_delta);

#ifdef NO_RADIUS_SCALE
	float light_scale = 1.0;
#else
	float light_scale = max(1.0 - (dist - light_radius.y)/(light_radius.x - light_radius.y), 0.0);
#endif

	vec3 dir_to_light =
#ifdef IS_ORTHO
	-spot_dir_view[0];
#else
 	light_delta/dist;
#endif

#ifndef NO_NORM_BUFFER
	vec3 view_normal = world_to_view * normal;

#	ifdef USE_SHADOWMAP
	if (dot(view_normal, dir_to_light) <= 0.0)
		light_scale = 0.0;
#	endif

#endif

#ifdef OMNI
	if (light_scale > 0.0)
#else
	// spotlight
	if (light_scale > 0.0 && light_radius.z > -1.0)
#endif
	{
		int faceIdx = 0;
		float normal_bias_scale = 1.0;

#ifdef USE_CASCADES
		// pick best face...
		int faceIdx2 = 0;
		float faceBlend = 0.0;

		for (int c=0; c<num_cascades; c++)
		{
			vec4 light_projected_uvs = spotlight_projection_matrix[c] * vec4(viewPos, 1.0);
			vec3 projuv = light_projected_uvs.xyz / light_projected_uvs.w;

			if (projuv.x >= 0.0 && projuv.y >= 0.0 && projuv.x <= 1.0 && projuv.y <= 1.0)
			{
				faceIdx = faceIdx2 = c;

				if (c+1 < num_cascades)
				{
					vec4 light_projected_uvs2 = spotlight_projection_matrix[c+1] * vec4(viewPos, 1.0);
					vec3 projuv2 = light_projected_uvs2.xyz / light_projected_uvs2.w;

					if (projuv2.x >= 0.0 && projuv2.y >= 0.0 && projuv2.x <= 1.0 && projuv2.y <= 1.0)
					{
						faceIdx2 = c+1;
						faceBlend = compute_cascade_blend(projuv.xy);
					}
				}

				break;
			}
		}

#endif

#ifdef IS_ORTHO
		// disabled - this causes cascades to appear sliding...
		// on the other hand if the bias is too large, shadows don't match at sharp edges
		// so... right now we only use the 1st texels size
		normal_bias_scale = 0.025*spotlight_texel_size[0*faceIdx];
#endif

		vec3 viewPosBiased = viewPos
#ifndef NO_NORM_BUFFER
				// normal bias
#	ifdef IS_ORTHO
				// apply linear-depth based bias as well - gets rid of horrible acne far away
				+ view_normal*(0.06*normal_bias_scale + max(linear_depth, 1.0)*0.005)
#	else
				+ view_normal*(0.01*dist
#		ifdef USE_SHADOWMAP
				 / (float(textureSize(tex8[0], 0).x) * (1.0/512.0))
#		endif
				)
#	endif
#endif
		;

#ifdef OMNI
		// pick best face!
		float bestFaceDot = 0.0;
		// note: negated
		vec3 light_delta_biased = viewPosBiased - light_pos_view;

		for (int f=0; f<6; f++)
		{
			float fdot = dot(light_delta_biased, spot_dir_view[f]);

			if (fdot > bestFaceDot)
			{
				bestFaceDot = fdot;
				faceIdx = f;
			}
		}
#endif

#ifndef NO_SPOT_SCALE
		float spot = -dot(spot_dir_view[faceIdx], dir_to_light);

		if (spot < light_radius.z)
			light_scale = 0.0;
		else
		{
			if (spot < light_radius.w)
				light_scale *= (spot - light_radius.z) / (light_radius.w - light_radius.z);
#else
		{
#endif

			vec4 light_projected_uvs = spotlight_projection_matrix[faceIdx] * vec4(viewPosBiased, 1.0);
			vec3 projuv = light_projected_uvs.xyz
#ifndef IS_ORTHO
			/ light_projected_uvs.w
#endif
			;

#	ifdef USE_SHADOWMAP

#		if !defined(NO_NORM_BUFFER) && !defined(IS_ORTHO)
			// disabled as it crippled large omnis..
			float ref_bias = 0.0;
			float ref_depth = projuv.z - ref_bias;
			// this isn't good either - gets rid of acne, but way too aggressive
			#ifndef OMNI
			ref_depth *= mix(1.0/1.003, 1.0, clamp(dist * 0.05, 0.0, 1.0));
			#endif
#		else
			// apply simple dumb bias - still get acne though!
			float ref_depth = projuv.z
#			ifdef IS_ORTHO
				// apply fixed bias, assuming linear mapping for ortho. helps get rid of sw mesa artifacts
				-0.5/65536.0
#			endif
#			if !defined(NO_SCALED_BIAS)
				* (1.0/1.001)
#			endif
			;
#		endif

			float shadow_tex_size_scale = 2.0/float(textureSize(tex8[0], 0).x);
			float shadow_scale = 0.0;

#ifdef USE_CASCADES
	// FIXME: copy-pasta!!!
			vec4 light_projected_uvs2 = spotlight_projection_matrix[faceIdx2] * vec4(viewPosBiased, 1.0);
			vec3 projuv2 = light_projected_uvs2.xyz
#ifndef IS_ORTHO
			/ light_projected_uvs2.w
#endif
			;

	#ifdef USE_SHADOWMAP

#		if !defined(NO_NORM_BUFFER) && !defined(IS_ORTHO)
			float ref_depth2 = projuv2.z - ref_bias;
#		else
			// apply simple dumb bias - still get acne though!
			float ref_depth2 = projuv2.z
#			ifdef IS_ORTHO
				// apply fixed bias, assuming linear mapping for ortho. helps get rid of sw mesa artifacts
				-0.5/65536.0
#			endif
#			if !defined(NO_SCALED_BIAS)
				* (1.0/1.001)
#			endif
			;
#		endif

			float shadow_scale2 = 0.0;
#	endif
#endif

			// simple brute-force 5-sample PCF
#ifndef PCF_16
			ivec2 iv = ivec2(gl_FragCoord.xy) & ivec2(3);
			float dith = dither[iv.y][iv.x];
			vec2 ofs = vec2(dith, dither[iv.x][iv.y]) * 1.0;
#else
			vec2 ofs = vec2(0.0);
#endif

			// this thing is necessary because GLSL (until 4.60 or so?)
			// doesn't allow to sample from array using a variable
#define PCF_LOOP(idx) \
	else if (faceIdx == idx) \
	{ \
		for (int i=0; i<PCF_SAMPLES; i++) \
		{ \
			vec2 shadow_uv = projuv.xy + (samples[i]*PCF_SAMPLE_SCALE + ofs)*shadow_tex_size_scale; \
			vec3 shadow_uvw = vec3(shadow_uv, ref_depth); \
			float shadow_pcf = texture(tex8[idx], shadow_uvw); \
			shadow_scale += shadow_pcf*(1.0/float(PCF_SAMPLES)); \
		} \
	}

#define PCF_LOOP2(idx) \
	else if (faceIdx2 == idx) \
	{ \
		for (int i=0; i<PCF_SAMPLES; i++) \
		{ \
			vec2 shadow_uv = projuv2.xy + (samples[i]*PCF_SAMPLE_SCALE + ofs)*shadow_tex_size_scale; \
			vec3 shadow_uvw = vec3(shadow_uv, ref_depth2); \
			float shadow_pcf = texture(tex8[idx], shadow_uvw); \
			shadow_scale2 += shadow_pcf*(1.0/float(PCF_SAMPLES)); \
		} \
	}

#if defined(USE_CASCADES)
			if (false);
			PCF_LOOP(0)
			PCF_LOOP(1)
			PCF_LOOP(2)
			PCF_LOOP(3)
			PCF_LOOP(4)
			PCF_LOOP(5)

			if (faceBlend == 0.0);
			PCF_LOOP2(0)
			PCF_LOOP2(1)
			PCF_LOOP2(2)
			PCF_LOOP2(3)
			PCF_LOOP2(4)
			PCF_LOOP2(5)

			shadow_scale = mix(shadow_scale, shadow_scale2, faceBlend);
#elif defined(OMNI)
			if (false);
			PCF_LOOP(0)
			PCF_LOOP(1)
			PCF_LOOP(2)
			PCF_LOOP(3)
			PCF_LOOP(4)
			PCF_LOOP(5)
#else
			for (int i=0; i<PCF_SAMPLES; i++)
			{
				vec2 shadow_uv = projuv.xy + (samples[i]*PCF_SAMPLE_SCALE + ofs)*shadow_tex_size_scale;
				vec3 shadow_uvw = vec3(shadow_uv, ref_depth);
				float shadow_pcf = texture(tex8[0], shadow_uvw);

				shadow_scale += shadow_pcf*(1.0/float(PCF_SAMPLES));
			}
#endif

			light_scale *= shadow_scale;
	#endif

#ifndef NO_SPEC_BUFFER
			specular.xyz *= light_scale;
#endif

#ifndef NO_PROJECTED_TEXTURE
			diffuse.xyz *= texture(tex, projuv.xy).xyz;
#else
#ifndef OMNI
			// need to reproject again due to normal bias applied above => should come up with something else...
			vec4 lproj = spotlight_projection_matrix[faceIdx] * vec4(viewPos, 1.0);
			vec3 lprojuv = lproj.xyz / lproj.w;

			if (lprojuv.x < 0.0 || lprojuv.x > 1.0 || lprojuv.y < 0.0 || lprojuv.y > 1.0)
				diffuse.xyz = vec3(0.0);
#endif
#endif

#ifdef CASCADE_DEBUG
			diffuse.xyz = vec3(0.0);
			diffuse[faceIdx % 3] = 1.0;

			if (faceIdx > 2)
				diffuse[faceIdx/3 % 3] = 1.0;
			light_scale = 1.0;
#endif
		}
	}

#ifndef NO_NORM_BUFFER
	float lscl = dlight_lambert(view_normal, dir_to_light);
#	ifdef FLASHLIGHT
	lscl = dlight_flashlight(lscl);
#	endif
	light_scale *= lscl;
#endif

	vec3 final = light_color.xyz * (light_scale*2.0) * diffuse;//light_color * light_scale;;

#ifndef NO_SPEC_BUFFER
#ifndef NO_NORM_BUFFER
	// specular:
	if (dot(specular.xyz, specular.xyz) > 0.0 && light_scale > 0.0)
	{
		vec3 reflectDir = reflect(dir_to_light, view_normal);

		// note: clamp should be faster than max here (analyzer)
		float specCos = clamp(dot(reflectDir, normalize(viewPos)), 0.0, 1.0);

		float specular_power = specular.w*128.0 + 0.5;

		float specFactor = pow(specCos, specular_power) * sqrt(light_scale);

		vec3 specColor = specular.xyz*light_color.xyz*(specFactor);

		final.xyz += specColor;
	}
#endif
#endif

#ifdef USE_AMBIENT
	vec3 ad = normal;

	// note: should be -r.z, but actually this is how my cubemap "screenshot" works
	vec3 ambDir = vec3(-ad.y, ad.z, ad.x);

	vec4 ambColor = textureLod(tex6, ambDir, 0.0);
	final.xyz += diffuse.xyz * ambColor.xyz * ambientScale;
#endif

	fragColor.xyz = final;
	fragColor.w = 0.0;
}
