#version 140

#ifdef GL_ES
precision highp float;
#endif

#include "../include/half_lambert.frag"
#include "../include/constants.vert"
#include "../include/clip_plane.frag"

flat in int instanceID;

out vec4 fragColor;

// deferred inputs
// deferred_diffuse
uniform sampler2D tex2;

#ifndef NO_NORM_BUFFER
// deferred_normal
uniform sampler2D tex3;
#endif

#ifndef NO_SPEC_BUFFER
// deferred_specular
uniform sampler2D tex4;
#endif

// depth
uniform sampler2D tex5;

// xy = positive 1/screen dims
// zw = projection matrix components
uniform vec4 ofs;

// these two specify frustum corners
uniform vec3 corner;
uniform vec3 cornerDelta;

uniform mat3 world_to_view;

uniform vec3 light_pos_view[MAX_LIGHT_INSTANCES];
uniform vec4 light_radius[MAX_LIGHT_INSTANCES];
uniform vec4 light_color[MAX_LIGHT_INSTANCES];

float LinearDepth(float d)
{
	// now with pre-baked denormalization and sign flip
	return ofs.w / (ofs.z + d);
}

void main()
{
	clip_pixel();

	vec2 uv = gl_FragCoord.xy * ofs.xy;
	float storedDepth = texture(tex5, uv).x;

	vec3 diffuse = texture(tex2, uv).xyz;

#ifndef NO_NORM_BUFFER
	// unpack world space normal
	vec3 normal = texture(tex3, uv).xyz;
	normal *= 2.0;
	normal -= vec3(1.0);
	// renormalization is necessary to reduce quantization errors
	normal = normalize(normal);
#endif

#ifndef NO_SPEC_BUFFER
	vec4 specular = texture(tex4, uv);
#endif

	vec3 viewPos = vec3(0.0);

	float linear_depth;

	linear_depth = LinearDepth(storedDepth);
	viewPos = (corner + gl_FragCoord.xyz*cornerDelta)*linear_depth;

	vec3 light_delta = viewPos - light_pos_view[instanceID];

	float dist = length(light_delta);

	vec4 radius = light_radius[instanceID];
	float light_scale = max(1.0 - (dist - radius.y)/(radius.x - radius.y), 0.0);

#ifndef NO_NORM_BUFFER
	vec3 dir_to_light = normalize(-light_delta);

	vec3 view_normal = world_to_view * normal;

	light_scale *= dlight_lambert(view_normal, dir_to_light);
#endif

	vec3 final = light_color[instanceID].xyz * (light_scale*2.0) * diffuse;

#ifndef NO_SPEC_BUFFER
#ifndef NO_NORM_BUFFER
	// specular:
	if (dot(specular.xyz, specular.xyz) > 0.0 && light_scale > 0.0)
	{
		vec3 reflectDir = reflect(dir_to_light, view_normal);

		// note: clamp should be faster than max here (analyzer)
		float specCos = clamp(dot(reflectDir, normalize(viewPos)), 0.0, 1.0);

		float specular_power = specular.w*128.0 + 0.5;

		float specFactor = pow(specCos, specular_power) * sqrt(light_scale);

		vec3 specColor = specular.xyz*light_color[instanceID].xyz*(specFactor);

		final.xyz += specColor;
	}
#endif
#endif

	fragColor.xyz = final;
	fragColor.w = 0.0;
}
