#include "half_lambert.frag"

uniform vec4 dynamic_light_positions[4];
uniform vec4 dynamic_light_colors[4];
// x = radius, y = inner radius
uniform vec4 dynamic_light_radii[4];

// global flashlight projected texture
uniform sampler2D tex7;
// global flashlight shadowmap
uniform
#ifdef GL_ES
highp
#endif
sampler2DShadow tex8;

// flashlight globals
uniform vec3 gflash_light_pos_view;
uniform vec4 gflash_light_radius;
uniform vec4 gflash_light_color;
uniform vec3 gflash_spot_dir_view;
// passing as vec4 array due to limitations in my material system
uniform vec4 gflash_projection_matrix[4];

// assume viewspace position/normal
vec3 trans_compute_flashlight(vec3 pos, vec3 norm)
{
	vec4 light_radius = gflash_light_radius;

	// early out if disabled
	if (light_radius.z <= -1.0)
		return vec3(0.0);

	vec3 light_pos_view = gflash_light_pos_view;

	vec3 light_delta = light_pos_view - pos;

	float dist = length(light_delta);

	if (dist < 0.0)
		return vec3(0.0);

	mat4 spotlight_projection_matrix;
	spotlight_projection_matrix[0] = gflash_projection_matrix[0];
	spotlight_projection_matrix[1] = gflash_projection_matrix[1];
	spotlight_projection_matrix[2] = gflash_projection_matrix[2];
	spotlight_projection_matrix[3] = gflash_projection_matrix[3];

	vec4 light_color = gflash_light_color;
	vec3 spot_dir_view = gflash_spot_dir_view;

	float light_scale = max(1.0 - ((dist - light_radius.y)/(light_radius.x - light_radius.y)), 0.0);
	float close_scale = dist * (1.0/0.5);
	light_scale *= min(1.0, close_scale*close_scale);

	vec3 dir_to_light = light_delta/dist;

	float spot = -dot(spot_dir_view, dir_to_light);

	// culled?
	if (spot < light_radius.z)
		return vec3(0.0);

	if (spot < light_radius.w)
		light_scale *= (spot - light_radius.z) / (light_radius.w - light_radius.z);

	vec4 light_projected_uvs = spotlight_projection_matrix * vec4(pos, 1.0);
	vec3 projuv = light_projected_uvs.xyz / light_projected_uvs.w;

	// shadow map

	vec2 samples[5]=vec2[5](
	vec2(-0.973062, -0.328256),
	vec2(0.110481, 0.712614),
	vec2(-0.818603, 0.638871),
	vec2(-0.018864, -0.385902),
	vec2(-0.430052, 0.142593)
	);

	mat4 dither = mat4(
	0.0/16.0-0.5, 8.0/16.0-0.5, 2.0/16.0-0.5, 10.0/16.0-0.5,
	12.0/16.0-0.5, 4.0/16.0-0.5, 14.0/16.0-0.5, 6.0/16.0-0.5,
	3.0/16.0-0.5, 11.0/16.0-0.5, 1.0/16.0-0.5, 9.0/16.0-0.5,
	15.0/16.0-0.5, 7.0/16.0-0.5, 13.0/16.0-0.5, 5.0/16.0-0.5);

	// apply simple dumb bias - still get acne though!
	float ref_depth = projuv.z * (1.0/1.001);

	float shadow_tex_size_scale = 2.0/float(textureSize(tex8, 0).x);
	float shadow_scale = 0.0;

	// simple brute-force 5-sample PCF

	ivec2 iv = ivec2(mod(gl_FragCoord.xy, vec2(4.0)));
	float dith = dither[iv.y][iv.x];
	vec2 ofs = vec2(dith, dither[iv.x][iv.y]);

	// 5 sample PCF (+hw 2x2) seems enough with dithering
	for (int i=0; i<5; i++)
	{
		vec2 shadow_uv = projuv.xy + (samples[i]*1.0 + ofs)*shadow_tex_size_scale;
		vec3 shadow_uvw = vec3(shadow_uv, ref_depth);
		float shadow_pcf = texture(tex8, shadow_uvw);

		shadow_scale += shadow_pcf/float(5.0);
	}

	light_scale *= shadow_scale;

	// shadow map end

	vec3 diffuse = texture(tex7, projuv.xy).xyz;
	light_scale *= dlight_lambert(norm, dir_to_light);

	return light_color.xyz * (light_scale*2.0) * diffuse;
}

vec3 trans_compute_dynamic_light(vec4 lpos, vec4 lcol, vec4 lrad, vec3 pos, vec3 norm)
{
	vec3 res = vec3(0.0);

	float radius = lrad.x;
	float innerRadius = lrad.y;

	if (radius <= 0.0)
		return res;

	vec3 dir_to_light = lpos.xyz - pos;

	float dist = length(dir_to_light);
	float intensity = max(1.0 - (dist - innerRadius) / (radius - innerRadius), 0.0);

	intensity *= max(dot(norm, dir_to_light), 0.0);

	return (lcol*intensity).xyz;
}

vec3 trans_compute_dynamic_lights(vec3 pos, vec3 norm)
{
	int i;
	vec3 res = vec3(0.0);

	for (i=0; i<4; i++)
	{
		res += trans_compute_dynamic_light(
			dynamic_light_positions[i],
			dynamic_light_colors[i],
			dynamic_light_radii[i],
			pos,
			norm);
	}

	res += trans_compute_flashlight(pos, norm);

	return res;
}
